// ********************
// * PlayerLog Plugin *
// ********************
// version 1.12
// by Pierre-Marie Baty <pm@bots-united.com>
// and AustinBotServer <austinbotserver@mindspring.com>

// LAST MODIFIED BY PM (change this line when you modify the CPP file)

#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>
#include <time.h>


// server population and chat log files
#define FILE_POPULATIONLOGFILE "cstrike/serverpopulation.log"
#define FILE_CHATLOGFILE "cstrike/serverchat.log"


enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable;
cvar_t logpopulation = {"logpopulation", "1", FCVAR_EXTDLL};
cvar_t logchat = {"logchat", "1", FCVAR_EXTDLL};
time_t rawtime;
char formattedtime[32];

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
void ClientDisconnect (edict_t *pEntity);
void ClientPutInServer (edict_t *pEntity);
void ClientCommand (edict_t *pEntity);
void InitLogFiles (void);
void LogToFile (const char *filename, const char *fmt, ...);


// START of Metamod stuff

enginefuncs_t meta_engfuncs;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   GetEntityAPI2, // pfnGetEntityAPI2()
   NULL, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   NULL, // pfnGetEngineFunctions()
   NULL, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "PlayerLog", // plugin name
   "1.12", // plugin version
   "22/05/2003", // date of creation
   "Pierre-Marie Baty <pm@bots-united.com> & AustinBotServer <austinbotserver@mindspring.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "PLAYERLOG", // plugin logtag
   PT_CHANGELEVEL, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   // initialize the plugin log files
   InitLogFiles ();

   // ask the engine to register the CVARs this plugin uses
   CVAR_REGISTER (&logpopulation);
   CVAR_REGISTER (&logchat);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEntityAPI2 (DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion)
{
   // this is one of the initialization functions hooked by metamod in the gamedll API

   // keep track of pointers to the functions we use in the gamedll API
   gFunctionTable.pfnClientDisconnect = ClientDisconnect;
   gFunctionTable.pfnClientPutInServer = ClientPutInServer;
   gFunctionTable.pfnClientCommand = ClientCommand;

   // copy the whole table for metamod to know which functions we are using here
   memcpy (pFunctionTable, &gFunctionTable, sizeof (DLL_FUNCTIONS));
   return (TRUE); // alright
}

// END of Metamod stuff




#ifdef _WIN32
int WINAPI DllMain (HINSTANCE hinstDLL, unsigned long fdwReason, LPVOID lpvReserved)
{
   // required Win32 DLL entry point. Good place for freeing stuff upon DLL unloading (in case
   // fdwReason would be DLL_PROCESS_DETACH) if needed. The same for Linux would be implemented
   // in an optional _fini() function.

   return (TRUE); // nothing to do here
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   // this function is called by the engine to tell the MOD DLL (hence the plugins too) the
   // address of its function table and its server global variables structure.

   // keep track of the engine functions and the server globals for our own use
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;
}


void ClientDisconnect (edict_t *pEntity)
{
   // this function is called whenever a client is VOLUNTARILY disconnected from the server,
   // either because the client dropped the connection, or because the server dropped him from
   // the game (latency timeout). The effect is the freeing of a client slot on the server. Note
   // that clients and bots disconnected because of a level change NOT NECESSARILY call this
   // function, because in case of a level change, it's a server shutdown, and not a normal
   // disconnection. Note that we can only hook player disconnections, because bots call
   // directly the right function in the GameDLL, without passing through the metamod API,
   // *UNLESS* the disconnection is provoked by the engine, either because of a mapchange, or
   // because of a server command ; thus we need to ensure the entity disconnecting is really
   // a human player, before logging him. Remember the infrastructure:
   // Engine DLL <===> metamod + plugin <===> bot DLL <===> Game DLL

   // don't care about bots
   if (pEntity->v.flags & FL_FAKECLIENT)
      RETURN_META (MRES_IGNORED);

   time (&rawtime); // get the system date and time
   sprintf (formattedtime, ctime (&rawtime)); // format the raw time
   formattedtime[strlen (formattedtime) - 1] = 0; // get rid of the trailing \n

   // if logging is enabled, log the date and the name of the player who left the game
   if (CVAR_GET_FLOAT ("logpopulation") > 0)
      LogToFile (FILE_POPULATIONLOGFILE, "[%s] %s (WONId %d) has left the game\n",
                 formattedtime, STRING (pEntity->v.netname), g_engfuncs.pfnGetPlayerWONId (pEntity));

   RETURN_META (MRES_IGNORED);
}


void ClientPutInServer (edict_t *pEntity)
{
   // this function is called once a just connected client actually enters the server, after
   // having downloaded and synchronized its resources with the of the server's. It's the
   // perfect place to hook for client connecting, since a client can always try to connect
   // passing the ClientConnect() step, and not be allowed by the server later (because of a
   // latency timeout or whatever reason). Note that we can only hook player connections,
   // because bots call directly the right function in the GameDLL, without passing through
   // the metamod API. I'm not sure whether the engine might forcibly decide to put itself a
   // bot client in the server, thus I check for the nature of the client being put. Remember
   // the infrastructure:
   // Engine DLL <===> metamod + plugin <===> bot DLL <===> Game DLL

   // don't care about bots
   if (pEntity->v.flags & FL_FAKECLIENT)
      RETURN_META (MRES_IGNORED);

   time (&rawtime); // get the system date and time
   sprintf (formattedtime, ctime (&rawtime)); // format the raw time
   formattedtime[strlen (formattedtime) - 1] = 0; // get rid of the trailing \n

   // if logging is enabled, log the date and the name of the player who joined the game
   if (CVAR_GET_FLOAT ("logpopulation") > 0)
      LogToFile (FILE_POPULATIONLOGFILE, "[%s] %s (WONId %d) joined the game\n",
                 formattedtime, STRING (pEntity->v.netname), g_engfuncs.pfnGetPlayerWONId (pEntity));

   RETURN_META (MRES_IGNORED);
}


void ClientCommand (edict_t *pEntity)
{
   // this function is called whenever the client whose player entity is pEntity issues a client
   // command. How it works is that clients all have a global string in their client DLL that
   // stores the command string; if ever that string is filled with characters, the client DLL
   // sends it to the engine as a command to be executed. When the engine has executed that
   // command, that string is reset to zero. By the server side, we can access this string
   // by asking the engine with the CmdArgv(), CmdArgs() and CmdArgc() functions that work just
   // like executable files argument processing work in C (argc gets the number of arguments,
   // command included, args returns the whole string, and argv returns the wanted argument
   // only). Here is a good place to hook for client commands such as "say" and "say_team",
   // which are the commands used for clients to send chat strings to each other.

   // don't care about bots
   if (pEntity->v.flags & FL_FAKECLIENT)
      RETURN_META (MRES_IGNORED);

   // do we care about logging player chat ?
   if (CVAR_GET_FLOAT ("logchat") > 0)
   {
      // is it a "say" client command (this player is chatting with all players) ?
      if (strcmp (CMD_ARGV (0), "say") == 0)
      {
         // then log the chat string to the chat log file, along with the player name and WONId
         LogToFile (FILE_CHATLOGFILE, "[%s] %s (WONId %d) says: %s\n",
                    formattedtime, STRING (pEntity->v.netname), g_engfuncs.pfnGetPlayerWONId (pEntity), CMD_ARGV (1));
      }

      // else is it a "say_team" client command (this player is chatting to his team only) ?
      else if (strcmp (CMD_ARGV (0), "say_team") == 0)
      {
         // then log the chat string to the chat log file, along with the player name and WONId
         LogToFile (FILE_CHATLOGFILE, "[%s] %s (WONId %d) says (TEAM-ONLY): %s\n",
                    formattedtime, STRING (pEntity->v.netname), g_engfuncs.pfnGetPlayerWONId (pEntity), CMD_ARGV (1));
      }
   }

   RETURN_META (MRES_IGNORED);
}


void InitLogFiles (void)
{
   // this function initializes the log file, creating one if none is found. It is meant to be
   // called when the server boots up.

   FILE *fp;

   // do we want to log the server population ?
   if (CVAR_GET_FLOAT ("logpopulation") > 0)
   {
      // test if the file exists, if it cannot be open for reading, then it needs to be created
      fp = fopen (FILE_POPULATIONLOGFILE, "r"); // open the file for reading
      if (fp == NULL)
      {
         // the file was not found ; so create a new one
         fp = fopen (FILE_POPULATIONLOGFILE, "w"); // create the log file in ASCII write mode
         fprintf (fp, "PlayerLog server population log file started\n\n"); // dump an init string
         fclose (fp); // close the file
      }
   }

   // do we want to log the server chat ?
   if (CVAR_GET_FLOAT ("logchat") > 0)
   {
      // test if the file exists, if it cannot be open for reading, then it needs to be created
      fp = fopen (FILE_CHATLOGFILE, "r"); // open the file for reading
      if (fp == NULL)
      {
         // the file was not found ; so create a new one
         fp = fopen (FILE_CHATLOGFILE, "w"); // open the log file in ASCII write mode, discard content
         fprintf (fp, "PlayerLog server chat log file started\n\n"); // dump an init string
         fclose (fp); // close the file
      }
   }

   return; // and return
}


void LogToFile (const char *filename, const char *fmt, ...)
{
   // this function logs a message to the message log file somewhere in the cstrike directory

   FILE *fp;
   va_list argptr;
   static char string[1024];

   // concatenate all the arguments in one string
   va_start (argptr, fmt);
   vsprintf (string, fmt, argptr);
   va_end (argptr);

   fp = fopen (filename, "a"); // open the log file in ASCII append mode, keep content
   fprintf (fp, string); // dump the string into the file
   fclose (fp); // close back the file

   return; // and return
}
